// generated by verigenSV.pl on Fri Jun  4 11:18:59 2021
// input file: mesi.cfg

// compiler directives
/* verilator lint_off CASEINCOMPLETE */
/* verilator lint_off DECLFILENAME */
/* verilator lint_off UNUSED */
/* verilator lint_off STMTDLY */
/* verilator lint_off UNOPTFLAT */


module MESI (
	input logic clk,
	input logic res_n,
	input logic [31:0] addr,
	input logic cp,
	input logic rd,
	input logic wr,
	output logic [3:0] mesi, 
	output logic [5:0] index, 
	output logic err );

  //
 // parameter constants 
//---------------------------
parameter ZERO = 1'b0;
parameter ONE = 1'b1;
parameter ZERO_MESI = 4'b0000;
parameter ZERO_TAG = 24'h000000;
parameter ZERO_ADDR = 6'b000000;
parameter MODIFIED = 4'b1000;
parameter EXCLUSIVE = 4'b0100;
parameter SHARED = 4'b0010;
parameter INVALID = 4'b0001;


  //
 // state encoding typedef 
//--------------------------------
typedef enum logic [2:0] {
	ID = 3'b000,
	RD = 3'b001,
	WR = 3'b010,
	rRD = 3'b101,
	rWR = 3'b110
} states_t;


// internal state variables
states_t STATE, NEXT_STATE;

  //
 // memory instantiations 
//---------------------------
RAM #( 4, 64 ) MESI_RAM(
	.clk(clk), .res_n(res_n), 
	.wr(MESI_RAM_wr), .wr_addr(MESI_RAM_wr_addr[5:0]), .wr_data(MESI_RAM_wr_data[3:0]), 
	.rd(MESI_RAM_rd), .rd_addr(MESI_RAM_rd_addr[5:0]), .rd_data(MESI_RAM_rd_data[3:0]) );

RAM #( 24, 64 ) TAG_RAM(
	.clk(clk), .res_n(res_n), 
	.wr(TAG_RAM_wr), .wr_addr(TAG_RAM_wr_addr[5:0]), .wr_data(TAG_RAM_wr_data[23:0]), 
	.rd(TAG_RAM_rd), .rd_addr(TAG_RAM_rd_addr[5:0]), .rd_data(TAG_RAM_rd_data[23:0]) );

   //
  // implicit continuous assignments 
 // (derived from <expressions> section) 
//----------------------------------
uwire IDLE; assign IDLE = ~res_n | (~rd & ~wr);
uwire IS_EXCLUSIVE; assign IS_EXCLUSIVE = POP_MESI_RAM(MESI_READ,addr[7:2])==EXCLUSIVE;
uwire IS_INVALID; assign IS_INVALID = POP_MESI_RAM(MESI_READ,addr[7:2])==INVALID;
uwire IS_MODIFIED; assign IS_MODIFIED = POP_MESI_RAM(MESI_READ,addr[7:2])==MODIFIED;
uwire IS_SHARED; assign IS_SHARED = POP_MESI_RAM(MESI_READ,addr[7:2])==SHARED;
uwire MESI_READ; assign MESI_READ = WRITE | (TAG_MATCH & (rREAD | rWRITE));
uwire MESI_WRITE; assign MESI_WRITE = READ | WRITE | (TAG_MATCH & (rREAD | rWRITE));
uwire READ; assign READ = res_n & cp & rd;
uwire TAG_MATCH; assign TAG_MATCH = POP_TAG_RAM(TAG_READ,addr[7:2])==addr[31:8];
uwire TAG_READ; assign TAG_READ = READ | WRITE | rREAD | rWRITE;
uwire TAG_WRITE; assign TAG_WRITE = !TAG_MATCH & READ;
uwire WRITE; assign WRITE = res_n & cp & wr;
uwire rREAD; assign rREAD = res_n & ~cp & rd;
uwire rWRITE; assign rWRITE = res_n & ~cp & wr;

  //
 // initials for simulator
//---------------------------
initial begin
  STATE = ID;

  // dump output (Verilator)
  if ($test$plusargs("trace") != 0) begin
    $display("[%0t] Tracing to logs/vlt_dump.vcd...\n", $time);
    $dumpfile("logs/vlt_dump.vcd");
    $dumpvars();
  end
  $display("[%0t] Model running...\n", $time);

end

  //
 // state transfer
//---------------------------
always_ff @( posedge clk ) begin
  if( res_n )
    STATE <= NEXT_STATE;
  else
    STATE <= ID;

  // condition for $finish (Verilator)
  if( rd==1'b1 && wr==1'b1 ) begin
    $write("*-* All Finished *-*\n");
    $finish;
  end   

end

// AssertionExample:
//always_ff @ (posedge clk) begin
//   AssertionExample: assert (rd==1'b1 ^ wr==1'b1);
//end

// coverage analysis (Verilator)
cover property (@(posedge clk) rd==1'b1 && wr==1'b1);

always_comb begin

    //
   // pre-case (default) assignments 
  //----------------------------------
  err = ZERO;
  index = addr[7:2];
  mesi = ZERO_MESI;
  PUSH_MESI_RAM(1'b0,addr[7:2],ZERO_MESI);
  PUSH_TAG_RAM(1'b0,addr[7:2],addr[31:8]);

  unique case ( STATE )
  ID : begin
	if ( IDLE ) begin
	  NEXT_STATE = ID;
	end
	if ( READ & TAG_MATCH ) begin
	  NEXT_STATE = RD;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],EXCLUSIVE);
	  mesi=EXCLUSIVE;
	end
	if ( READ & ~TAG_MATCH ) begin
	  NEXT_STATE = RD;
	  PUSH_TAG_RAM(TAG_WRITE,addr[7:2],addr[31:8]);
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],EXCLUSIVE);
	  mesi=EXCLUSIVE;
	end
	if ( WRITE & TAG_MATCH ) begin
	  NEXT_STATE = WR;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],MODIFIED);
	  mesi=MODIFIED;
	end
	if ( WRITE & ~TAG_MATCH ) begin
	  NEXT_STATE = WR;
	  err=ONE;
	end
	if ( rWRITE & TAG_MATCH ) begin
	  NEXT_STATE = rWR;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],INVALID);
	  mesi=INVALID;
	end
	if ( rREAD & TAG_MATCH ) begin
	  NEXT_STATE = rRD;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],SHARED);
	  mesi=SHARED;
	end
	if ( rREAD & ~TAG_MATCH ) begin
	  NEXT_STATE = ID;
	end
  end // ID
  RD : begin
	if ( IDLE ) begin
	  NEXT_STATE = ID;
	end
	if ( READ & TAG_MATCH ) begin
	  NEXT_STATE = RD;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],EXCLUSIVE);
	  mesi=EXCLUSIVE;
	end
	if ( READ & ~TAG_MATCH ) begin
	  NEXT_STATE = RD;
	  PUSH_TAG_RAM(TAG_WRITE,addr[7:2],addr[31:8]);
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],EXCLUSIVE);
	  mesi=EXCLUSIVE;
	end
  end // RD
  WR : begin
	if ( IDLE ) begin
	  NEXT_STATE = ID;
	end
	if ( WRITE & TAG_MATCH ) begin
	  NEXT_STATE = WR;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],MODIFIED);
	  mesi=MODIFIED;
	end
	if ( WRITE & ~TAG_MATCH ) begin
	  NEXT_STATE = ID;
	  err=ONE;
	end
  end // WR
  rRD : begin
	if ( IDLE ) begin
	  NEXT_STATE = ID;
	end
	if ( rREAD & TAG_MATCH ) begin
	  NEXT_STATE = rRD;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],SHARED);
	  mesi=SHARED;
	end
	if ( rREAD & ~TAG_MATCH ) begin
	  NEXT_STATE = rRD;
	  PUSH_TAG_RAM(TAG_WRITE,addr[7:2],addr[31:8]);
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],SHARED);
	  mesi=SHARED;
	end
  end // rRD
  rWR : begin
	if ( IDLE ) begin
	  NEXT_STATE = ID;
	end
	if ( rWRITE & TAG_MATCH ) begin
	  NEXT_STATE = rWR;
	  PUSH_MESI_RAM(MESI_WRITE,addr[7:2],INVALID);
	  mesi=INVALID;
	end
	if ( rWRITE & ~TAG_MATCH ) begin
	  NEXT_STATE = rWR;
	end
  end // rWR
  endcase
end // always_comb


  //
 // memory functions
//---------------------------
logic MESI_RAM_rd;
logic [4-1:0] MESI_RAM_rd_data;
logic [$clog2(64)-1:0] MESI_RAM_rd_addr;

function automatic [4-1:0] POP_MESI_RAM( read, [$clog2(64)-1:0] read_addr);
   MESI_RAM_rd = read;
   MESI_RAM_rd_addr = read_addr;
   return MESI_RAM_rd_data;
endfunction

logic MESI_RAM_wr;
logic [4-1:0] MESI_RAM_wr_data;
logic [$clog2(64)-1:0] MESI_RAM_wr_addr;

function automatic void PUSH_MESI_RAM( write, [$clog2(64)-1:0] write_addr, [4-1:0] write_data);
   MESI_RAM_wr = write;
   MESI_RAM_wr_addr = write_addr;
   MESI_RAM_wr_data = write_data;
endfunction

function automatic void NOP_MESI_RAM( write, [$clog2(64)-1:0] write_addr, [4-1:0] write_data);
   MESI_RAM_rd = 0;
   MESI_RAM_wr = 0;
   MESI_RAM_wr_addr = '0;
   MESI_RAM_wr_data = '0;
endfunction

logic TAG_RAM_rd;
logic [24-1:0] TAG_RAM_rd_data;
logic [$clog2(64)-1:0] TAG_RAM_rd_addr;

function automatic [24-1:0] POP_TAG_RAM( read, [$clog2(64)-1:0] read_addr);
   TAG_RAM_rd = read;
   TAG_RAM_rd_addr = read_addr;
   return TAG_RAM_rd_data;
endfunction

logic TAG_RAM_wr;
logic [24-1:0] TAG_RAM_wr_data;
logic [$clog2(64)-1:0] TAG_RAM_wr_addr;

function automatic void PUSH_TAG_RAM( write, [$clog2(64)-1:0] write_addr, [24-1:0] write_data);
   TAG_RAM_wr = write;
   TAG_RAM_wr_addr = write_addr;
   TAG_RAM_wr_data = write_data;
endfunction

function automatic void NOP_TAG_RAM( write, [$clog2(64)-1:0] write_addr, [24-1:0] write_data);
   TAG_RAM_rd = 0;
   TAG_RAM_wr = 0;
   TAG_RAM_wr_addr = '0;
   TAG_RAM_wr_data = '0;
endfunction

endmodule


  //
 // memory
//---------------------------
module RAM ( clk, res_n, wr, wr_addr, wr_data, rd, rd_addr, rd_data );

parameter DATA_WIDTH = 32;
parameter NUM_ENTRIES = 16;
parameter ADDR_WIDTH = $clog2(NUM_ENTRIES);

input  uwire clk, res_n, rd, wr;
input  uwire [DATA_WIDTH-1:0] wr_data;
input  uwire [ADDR_WIDTH-1:0] rd_addr, wr_addr;
output uwire [DATA_WIDTH-1:0] rd_data;

parameter ZERO_ADDR = {ADDR_WIDTH{1'b0}};
parameter ZERO_DATA = {DATA_WIDTH{1'b0}};
parameter ONE_DATA = {DATA_WIDTH{1'b1}};
    
logic    [DATA_WIDTH-1:0] MEMORY [NUM_ENTRIES-1:0];

integer i;

initial begin
   // Clear RAM before simulation start
   for ( i = 0; i < NUM_ENTRIES; i = i + 1 ) begin
       MEMORY[ i ] = ZERO_DATA;
   end
end

// IN-CYCLE READ ACCESS
assign rd_data =  ( rd ) ? MEMORY[rd_addr] : 'Z;

always_ff @ ( negedge clk ) begin
    // WRITE ACCESS
   if( wr ) MEMORY[wr_addr] <= wr_data;
end

endmodule
